@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 
 """
-Copyright 2008-2014 Wolfgang Rohdewald <wolfgang@rohdewald.de>
+Copyright 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>
 
 Kajongg is free software you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -27,15 +27,18 @@ from uitile import UITile
 from meld import Meld, MeldList
 from hand import Hand
 from board import Board
+from sound import Sound
 
 from log import logDebug
-from common import Internal, Debug, isAlive
+from common import Internal, Debug, isAlive, StrMixin
 
 
-class TileAttr(object):
+class TileAttr(StrMixin):
 
     """a helper class for syncing the hand board, holding relevant
-    tile attributes"""
+    tile attributes.
+    xoffset and yoffset are expressed in number of tiles but may be
+    fractional for adding distances between melds"""
 
     def __init__(self, hand, meld=None, idx=None, xoffset=None, yoffset=None):
         if isinstance(hand, UITile):
@@ -52,7 +55,7 @@ class TileAttr(object):
             # dark and focusable are different in a ScoringHandBoard
             self.focusable = self.setFocusable(hand, meld, idx)
             if self.tile in Debug.focusable:
-                logDebug(u'TileAttr %s:%s' % (self.tile, self.focusable))
+                logDebug('TileAttr %s:%s' % (self.tile, self.focusable))
 
     def setDark(self):
         """should the tile appear darker?"""
@@ -78,9 +81,6 @@ class TileAttr(object):
              ' dark' if self.dark else '',
              ' focusable' if self.focusable else ''))
 
-    def __repr__(self):
-        return 'TileAttr(%s)' % str(self)
-
 
 class HandBoard(Board):
 
@@ -95,7 +95,7 @@ class HandBoard(Board):
         self.exposedMeldDistance = 0.15
         self.concealedMeldDistance = 0.0
         self.lowerY = 1.0
-        Board.__init__(self, 15.6, 2.0, Tileset.activeTileset())
+        Board.__init__(self, 15.6, 2.0, Tileset.current())
         self.isHandBoard = True
         self.tileDragEnabled = False
         self.setParentItem(player.front)
@@ -191,27 +191,22 @@ class HandBoard(Board):
                 meldX += meldDistance
         return sorted(result, key=lambda x: x.yoffset * 100 + x.xoffset)
 
-    def newBonusPositions(self, bonusTiles, newTilePositions):
-        """returns list(TileAttr)
-        calculate places for bonus tiles. Put them all in one row,
-        right adjusted. If necessary, extend to the right even
-        outside of our board"""
-        positions = list(x.xoffset for x in newTilePositions if x.yoffset == 0)
-        upperLen = max(positions) if positions else 0
-        positions = list(x.xoffset for x in newTilePositions if x.yoffset != 0)
-        lowerLen = max(positions) if positions else 0
-# TODO: keep them in the row they are in as long as there is room
-        if upperLen < lowerLen:
-            bonusY = 0.0
-            tileLen = upperLen
-        else:
-            bonusY = self.lowerY
-            tileLen = lowerLen
-        tileLen += 1 + self.exposedMeldDistance
-        newBonusTiles = list(self.tileAttrClass(x) for x in bonusTiles)
-        xPos = 13 - len(newBonusTiles)
-        xPos = max(xPos, tileLen)
+    def placeBoniInRow(self, bonusTiles, tilePositions, bonusY, keepTogether=True):
+        """Try to place bonusTiles in upper or in lower row.
+        tilePositions are the normal tiles, already placed.
+        If there is no space, return None
+
+        returns list(TileAttr)"""
+        positions = list(x.xoffset for x in tilePositions if x.yoffset == bonusY)
+        rightmostTileX = max(positions) if positions else 0
+        placeBoni = bonusTiles[:]
+        while 13 - len(placeBoni) < rightmostTileX + 1 + self.exposedMeldDistance:
+            if keepTogether:
+                return list()
+            placeBoni = placeBoni[:-1]
         result = list()
+        xPos = 13 - len(placeBoni)
+        newBonusTiles = list(self.tileAttrClass(x) for x in placeBoni)
         for bonus in sorted(newBonusTiles, key=lambda x: x.tile.key):
             bonus.xoffset, bonus.yoffset = xPos, bonusY
             bonus.dark = False
@@ -219,9 +214,30 @@ class HandBoard(Board):
             xPos += 1
         return result
 
-    def calcPlaces(self, tiles):
-        """returns a dict. Keys are existing tiles, Values are TileAttr instances.
-        Values may be None: This is a tile to be removed from the board."""
+    def newBonusPositions(self, bonusTiles, newTilePositions):
+        """returns list(TileAttr)
+        calculate places for bonus tiles. Put them all in one row,
+        right adjusted. If necessary, extend to the right even
+        outside of our board"""
+        if not bonusTiles:
+            return list()
+        bonusTiles = sorted(bonusTiles, key=lambda x: hash(x.tile))
+        result = (
+            self.placeBoniInRow(bonusTiles, newTilePositions, 0.0)
+            or
+            self.placeBoniInRow(bonusTiles, newTilePositions, self.lowerY))
+        if not result:
+            # we cannot place all bonus tiles in the same row!
+            result = self.placeBoniInRow(bonusTiles, newTilePositions, 0.0, keepTogether=False)
+            result.extend(self.placeBoniInRow(
+                bonusTiles[len(result):], newTilePositions, self.lowerY, keepTogether=False))
+
+        assert len(bonusTiles) == len(result)
+        return result
+
+    def placeTiles(self, tiles):
+        """tiles are all tiles for this board.
+        returns a list of those uiTiles which are placed on the board"""
         oldTiles = dict()
         oldBonusTiles = dict()
         for uiTile in tiles:
@@ -267,7 +283,7 @@ class HandBoard(Board):
             uiTile.setBoard(self, newPos.xoffset, newPos.yoffset)
             uiTile.dark = newPos.dark
             uiTile.focusable = newPos.focusable
-        return result
+        return result.keys()
 
     def _avoidCrossingMovements(self, places):
         """not needed for all HandBoards"""
@@ -283,7 +299,7 @@ class HandBoard(Board):
         allTiles = self.uiTiles[:]
         if adding:
             allTiles.extend(adding)
-        self.calcPlaces(allTiles)
+        self.placeTiles(allTiles)
 
     def checkTiles(self):
         """does the logical state match the displayed tiles?"""
@@ -331,13 +347,13 @@ class PlayingHandBoard(HandBoard):
         allTiles = self.uiTiles[:]
         if adding:
             allTiles.extend(adding)
-        newPlaces = self.calcPlaces(allTiles)
-        source = adding if adding else newPlaces.keys()
+        newTiles = self.placeTiles(allTiles)
+        source = adding if adding else newTiles
         focusCandidates = list(x for x in source
                                if x.focusable and x.tile.isConcealed)
         if not focusCandidates:
             # happens if we just exposed a claimed meld
-            focusCandidates = list(x for x in newPlaces.keys()
+            focusCandidates = list(x for x in newTiles
                                    if x.focusable and x.tile.isConcealed)
         focusCandidates = sorted(focusCandidates, key=lambda x: x.xoffset)
         if focusCandidates:
@@ -443,3 +459,9 @@ class PlayingHandBoard(HandBoard):
         Internal.scene.discardBoard.discardTile(lastDiscard)
         for uiTile in self.uiTiles:
             uiTile.focusable = False
+
+    def addUITile(self, uiTile):
+        """add uiTile to this board"""
+        Board.addUITile(self, uiTile)
+        if uiTile.isBonus and not self.player.game.isScoringGame():
+            Sound.bonus()
